# coding: utf-8
from __future__ import unicode_literals

import os
from io import StringIO

from .compat import text
from .exceptions import InvalidBatch
from .utils import guess_mimetype

try:
    from typing import TYPE_CHECKING
    if TYPE_CHECKING:
        from typing import Any, BinaryIO, Dict, List, Optional, Text, Union
        from io import FileIO
        from .directories import API as DirectoriesAPI
        from .documents import API as DocumentsAPI
        from .endpoint import APIEndpoint
        from .groups import API as GroupsAPI
        from .operations import API as OperationsAPI
        from .tasks import API as TasksAPI
        from .uploads import API as UploadsAPI
        from .users import API as UsersAPI
        from .workflows import API as WorkflowsAPI
except ImportError:
    pass

""" Base classes """


class Model(object):
    """ Base class for all entities. """
    _valid_properties = {}  # type: Dict[Text, Any]
    service = None  # type: APIEndpoint
    uid = None  # type: Text

    def __init__(self, service=None, **kwargs):
        # type: (Optional[APIEndpoint], Any) -> None
        self.service = service

    def __repr__(self):
        # type: () -> Text
        attrs = ', '.join('{}={!r}'.format(
            attr, getattr(self, attr.replace('-', '_'), None))
            for attr in sorted(self._valid_properties))
        return '<{} {}>'.format(self.__class__.__name__, attrs)

    def as_dict(self):
        # type: () -> Dict[Text, Any]
        """ Returns a dict representation of the resource. """
        result = {}
        for key in self._valid_properties:
            val = getattr(self, key.replace('-', '_'))
            if val is None:
                continue
            # Parse lists of objects
            elif (isinstance(val, list) and len(val) > 0
                  and isinstance(val[0], Model)):
                val = [item.as_dict() for item in val]

            result[key] = val
        return result

    @classmethod
    def parse(cls, json, service=None):
        # type: (Dict[Text, Any], Optional[APIEndpoint]) -> Model
        """ Parse a JSON object into a model instance. """
        model = cls()

        if service:
            setattr(model, 'service', service)

        for key, val in json.items():
            if key in cls._valid_properties:
                key = key.replace('-', '_')
                setattr(model, key, val)
        return model

    def save(self):
        # type: () -> None
        """ Save the resource. """
        self.service.put(self)


class RefreshableModel(Model):

    def load(self, model=None):
        # type: (Optional[Union[Model, Dict[Text, Any]]]) -> None
        """
        Reload the Model.

        If model is not None, copy from its attributes,
        otherwise query the server for the entity with its uid.

        :param model: the entity to copy
        :return: the refreshed model
        """
        if not model:
            model = self.service.get(self.uid)

        if isinstance(model, dict):
            def get_prop(obj, key):
                return obj.get(key)
        else:
            def get_prop(obj, key):
                return getattr(obj, key)

        for key in self._valid_properties:
            key = key.replace('-', '_')
            setattr(self, key, get_prop(model, key))


""" Entities """


class Batch(Model):
    """ Upload batch. """

    _valid_properties = {
        'batchId': None,
        'dropped': None,
    }
    service = None  # type: UploadsAPI

    def __init__(self, **kwargs):
        # type: (Any) -> None
        super(Batch, self).__init__(**kwargs)
        self.batchId = None  # type: Text
        self.blobs = {}  # type: Dict[int, Blob]
        self._upload_idx = 0
        for key, default in Batch._valid_properties.items():
            setattr(self, key, kwargs.get(key, default))

    @property
    def uid(self):
        # type: () -> Text
        return self.batchId

    @uid.setter
    def uid(self, value):
        # type: (Text) -> None
        self.batchId = value

    def cancel(self):
        # type: () -> None
        """ Cancel an upload batch. """
        if not self.batchId:
            return
        self.service.delete(self.uid)
        self.batchId = None

    def delete(self, file_idx):
        """ Delete a blob from the batch. """
        if self.batchId:
            self.service.delete(self.uid, file_idx=file_idx)
            self.blobs[file_idx] = None

    def get(self, file_idx):
        # type: (int) -> Blob
        """
        Get the blob info.

        :param file_idx: the index of the blob in the batch
        :return: the corresponding blob
        """
        if self.batchId is None:
            raise InvalidBatch(
                'Cannot fetch blob for inexistant/deleted batch.')
        blob = self.service.get(self.uid, file_idx=file_idx)
        self.blobs[file_idx] = blob
        return blob

    def upload(self, blob, **kwargs):
        # type: (Blob, Any) -> Blob
        """
        Upload a blob.

        :param blob: the blob to upload
        :param kwargs: the upload settings
        :return: the blob info
        """
        blob = self.service.upload(self, blob, **kwargs)
        self.blobs[self._upload_idx] = blob
        self._upload_idx += 1
        return blob

    def execute(self, operation, file_idx=None, params=None):
        # type: (Text, int, Dict[Text, Any]) -> Any
        """
        Execute an operation on this batch.

        :param operation: operation to execute
        :param file_idx: target file of the operation
        :param params: parameters of the operation
        :return: the output of the operation
        """
        return self.service.execute(self, operation, file_idx, params)

    def attach(self, doc, file_idx=None):
        # type: (Text) -> Any
        """
        Attach one or all files of this batch to a document.

        :param doc: document to attach
        :param file_idx: target file
        :return: the output of the attach operation
        """
        return self.service.attach(self, doc, file_idx)


class Blob(Model):
    """ Blob superclass used for metadata. """

    _valid_properties = {
        'uploaded': 'true',
        'name': None,
        'uploadType': None,
        'size': 0,
        'uploadedSize': 0,
        'fileIdx': None,
        'mimetype': None,
        'uploadedChunkIds': [],
        'chunkCount': 0
    }
    service = None  # type: UploadsAPI

    def __init__(self, **kwargs):
        # type: (Any) -> None
        super(Blob, self).__init__(**kwargs)
        for key, default in Blob._valid_properties.items():
            if key == 'uploaded':
                val = kwargs.get(key, 'true') == 'true'
            elif key == 'size':
                val = kwargs.get(key, 0)
            elif key == 'uploadedSize':
                val = kwargs.get(key, kwargs.get('size', 0))
            else:
                val = kwargs.get(key, default)
            setattr(self, key, val)

    @classmethod
    def parse(cls, json, service=None):
        # type: (Dict[Text, Any], Optional[APIEndpoint]) -> Blob
        """ Parse a JSON object into a blob instance. """
        model = cls()

        if service:
            setattr(model, 'service', service)

        for key, val in json.items():
            if key in cls._valid_properties:
                setattr(model, key, val)

        if model.uploaded and model.uploadedSize == 0:
            model.uploadedSize = model.size
        return model

    def to_json(self):
        # type: () -> Dict[Text, Text]
        """ Return a JSON object used during the upload. """
        return {
            'upload-batch': self.batch_id,
            'upload-fileId': text(self.fileIdx),
        }


class BufferBlob(Blob):
    """
    InMemory content to upload to Nuxeo.

    Acts as a context manager so its data can be read
    with the `with` statement.
    """

    stringio = None  # type: Optional[StringIO]

    def __init__(self, data, **kwargs):
        # type: (Text, Any) -> None
        """
        :param data: content to upload to Nuxeo
        :param **kwargs: named attributes
        """
        super(BufferBlob, self).__init__(**kwargs)
        self.buffer = data
        self.mimetype = 'application/octet-stream'

    @property
    def data(self):
        # type: () -> StringIO
        """ Request data. """
        return self.stringio

    def __enter__(self):
        self.stringio = StringIO(self.buffer)
        return self.stringio

    def __exit__(self, *args):
        if self.stringio:
            self.stringio.close()


class FileBlob(Blob):
    """
    File to upload to Nuxeo.

    Acts as a context manager so its data can be read
    with the `with` statement.
    """

    # File descriptor
    fd = None  # type: Optional[BinaryIO]

    def __init__(self, path, **kwargs):
        # type: (Text, Any) -> None
        """
        :param path: file path
        :param **kwargs: named attributes
        """
        super(FileBlob, self).__init__(**kwargs)
        self.path = path
        self.name = os.path.basename(self.path)
        self.size = os.path.getsize(self.path)
        self.mimetype = (self.mimetype or
                         guess_mimetype(self.path))  # type: Text

    @property
    def data(self):
        # type: () -> BinaryIO
        """
        Request data.

        The caller has to close the file descriptor
        himself if he doesn't open it with the
        context manager.
        """
        return self.fd

    def __enter__(self):
        self.fd = open(self.path, 'rb')
        return self.fd

    def __exit__(self, *args):
        if self.fd:
            self.fd.close()


class Directory(Model):
    """ Directory. """
    _valid_properties = {
        'entity-type': 'directory',
        'directoryName': None,
        'entries': []
    }
    service = None  # type: DirectoriesAPI

    def __init__(self, **kwargs):
        # type: (Any) -> None
        super(Directory, self).__init__(**kwargs)
        for key, default in Directory._valid_properties.items():
            key = key.replace('-', '_')
            setattr(self, key, kwargs.get(key, default))

    @property
    def uid(self):
        # type: () -> Text
        return self.directoryName

    def get(self, entry):
        # type: (Text) -> DirectoryEntry
        """
        Get an entry of the directory.

        :param entry: the name of the entry
        :return: the corresponding directory entry
        """
        return self.service.get(self.uid, dir_entry=entry)

    def create(self, entry):
        # type: (DirectoryEntry) -> DirectoryEntry
        """ Create an entry in the directory. """
        return self.service.post(entry, dir_name=self.uid)

    def save(self, entry):
        # type: (DirectoryEntry) -> DirectoryEntry
        """ Save a modified entry of the directory. """
        return self.service.put(entry, dir_name=self.uid)

    def delete(self, entry=None):
        # type: (Text) -> None
        """
        Delete the directory or one of its entries.

        :param entry: if specified, the entry to delete
        """
        self.service.delete(self.uid, dir_entry=entry)

    def exists(self, entry):
        # type: (Text) -> bool
        """
        Check if an entry is in the directory.

        :param entry: the entry name
        :return: True if it exists, else False
        """
        return self.service.exists(self.uid, dir_entry=entry)


class DirectoryEntry(Model):
    """ Directory entry. """
    _valid_properties = {
        'entity-type': 'directoryEntry',
        'directoryName': None,
        'properties': {},
    }
    service = None  # type: DirectoriesAPI

    def __init__(self, **kwargs):
        # type: (Any) -> None
        super(DirectoryEntry, self).__init__(**kwargs)
        for key, default in DirectoryEntry._valid_properties.items():
            key = key.replace('-', '_')
            setattr(self, key, kwargs.get(key, default))

    @property
    def uid(self):
        # type: () -> Text
        return self.properties['id']

    def save(self):
        # type: () -> DirectoryEntry
        """ Save the entry. """
        return self.service.put(self, self.directoryName)

    def delete(self):
        # type: () -> None
        """ Delete the entry. """
        self.service.delete(self.directoryName, self.uid)


class Document(RefreshableModel):
    """ Document. """
    _valid_properties = {
        'entity-type': 'document',
        'repository': 'default',
        'name': None,
        'uid': None,
        'path': None,
        'type': None,
        'state': None,
        'parentRef': None,
        'versionLabel': None,
        'isCheckedOut': False,
        'isTrashed': False,
        'isVersion': False,
        'isProxy': False,
        'title': None,
        'lastModified': None,
        'properties': {},
        'facets': [],
        'changeToken': None,
        'contextParameters': {},
    }
    service = None  # type: DocumentsAPI

    def __init__(self, **kwargs):
        # type: (Any) -> None
        super(Document, self).__init__(**kwargs)
        for key, default in Document._valid_properties.items():
            key = key.replace('-', '_')
            setattr(self, key, kwargs.get(key, default))

    @property
    def workflows(self):
        # type: () -> List[Workflow]
        """ Return the workflows associated with the document. """
        return self.service.workflows(self)

    def add_permission(self, params):
        # type: (Dict[Text, Any]) -> None
        """
        Add a permission to a document.

        :param params: permission to add
        """
        return self.service.add_permission(self.uid, params)

    def convert(self, params):
        # type: (Dict[Text, Any]) -> Union[Dict[Text, Any], Text]
        """
        Convert the document to another format.

        :param params: Converter permission
        :return: the converter result
        """
        return self.service.convert(self.uid, params)

    def delete(self):
        # type: () -> None
        """ Delete the document. """
        self.service.delete(self.uid)

    def fetch_acls(self):
        # type: () -> Dict[Text, Any]
        """ Fetch document ACLs. """
        return self.service.fetch_acls(self.uid)

    def fetch_audit(self):
        # type: () -> Dict[Text, Any]
        """ Fetch audit for current document. """
        return self.service.fetch_audit(self.uid)

    def fetch_blob(self, xpath='blobholder:0'):
        # type: (Text) -> Blob
        """
        Retrieve one of the blobs attached to the document.

        :param xpath: the xpath to the blob
        :return: the blob
        """
        return self.service.fetch_blob(uid=self.uid, xpath=xpath)

    def fetch_lock_status(self):
        # type: () -> Dict[Text, Any]
        """ Get lock informations. """
        return self.service.fetch_lock_status(self.uid)

    def fetch_rendition(self, name):
        # type: (Text) -> Union[Text, bytes]
        """
        :param name: Rendition name to use
        :return: The rendition content
        """
        return self.service.fetch_rendition(self.uid, name)

    def fetch_renditions(self):
        # type: () -> List[Union[Text, bytes]]
        """
        :return: Available renditions for this document
        """
        return self.service.fetch_renditions(self.uid)

    def follow_transition(self, name):
        # type: (Text) -> None
        """
        Follow a lifecycle transition on this document.

        :param name: transition name
        """
        doc = self.service.follow_transition(self.uid, name)
        self.load(doc)

    def get(self, prop):
        # type: (Text) -> Any
        """ Get a property of the document by its name. """
        return self.properties[prop]

    def has_permission(self, permission):
        # type: (Text) -> bool
        """ Verify if a document has the permission. """
        return self.service.has_permission(self.uid, permission)

    def is_locked(self):
        # type: () -> bool
        """ Get the lock status. """
        return not not self.fetch_lock_status()

    def lock(self):
        # type: () -> Dict[Text, Any]
        """ Lock the document. """
        return self.service.lock(self.uid)

    def move(self, dst, name=None):
        # type: (Text, Optional[Text]) -> None
        """
        Move a document into another parent.

        :param dst: The new parent path
        :param name: Rename the document if specified
        """
        doc = self.service.move(self.uid, dst, name)
        self.load(doc)

    def remove_permission(self, params):
        # type: (Dict[Text, Any]) -> None
        """ Remove a permission to a document. """
        return self.service.remove_permission(self.uid, params)

    def set(self, properties):
        # type: (Dict[Text, Any]) -> None
        """ Add/update the properties of the document. """
        self.properties.update(properties)

    def trash(self):
        # type: () -> None
        doc = self.service.trash(self.uid)
        self.load(doc)

    def unlock(self):
        # type: () -> Dict[Text, Any]
        """ Unlock the document. """
        return self.service.unlock(self.uid)

    def untrash(self):
        # type: () -> None
        doc = self.service.untrash(self.uid)
        self.load(doc)


class Group(Model):
    """ User group. """
    _valid_properties = {
        'entity-type': 'group',
        'groupname': None,
        'grouplabel': None,
        'memberUsers': [],
        'memberGroups': [],
    }
    service = None  # type: GroupsAPI

    def __init__(self, **kwargs):
        # type: (Any) -> None
        super(Group, self).__init__(**kwargs)
        for key, default in Group._valid_properties.items():
            key = key.replace('-', '_')
            setattr(self, key, kwargs.get(key, default))

    @property
    def uid(self):
        # type: () -> Text
        return self.groupname

    def delete(self):
        """ Delete the group. """
        # type: () -> None
        self.service.delete(self.uid)


class Operation(Model):
    """ Automation operation. """
    _valid_properties = {
        'command': None,
        'input_obj': None,
        'params': {},
        'context': None,
        'progress': 0,
    }
    service = None  # type: OperationsAPI

    def __init__(self, **kwargs):
        # type: (Any) -> None
        super(Operation, self).__init__(**kwargs)
        for key, default in Operation._valid_properties.items():
            setattr(self, key, kwargs.get(key, default))

    def execute(self, **kwargs):
        # type: (Any) -> Any
        """ Execute the operation. """
        return self.service.execute(self, **kwargs)


class Task(RefreshableModel):
    """ Workflow task. """
    _valid_properties = {
        'entity-type': 'task',
        'id': None,
        'name': None,
        'workflowInstanceId': None,
        'workflowModelName': None,
        'state': None,
        'directive': None,
        'created': None,
        'dueDate': None,
        'nodeName': None,
        'targetDocumentIds': [],
        'actors': [],
        'comments': [],
        'variables': {},
        'taskInfo': {},
    }
    service = None  # type: TasksAPI

    def __init__(self, **kwargs):
        # type: (Any) -> None
        super(Task, self).__init__(**kwargs)
        for key, default in Task._valid_properties.items():
            key = key.replace('-', '_')
            setattr(self, key, kwargs.get(key, default))

    @property
    def uid(self):
        # type: () -> Text
        return self.id

    def complete(self, action, variables=None, comment=None):
        # type: (Text, Optional[Dict[Text, Any]], Optional[Text]) -> None
        """ Complete the action of a task. """
        updated_task = self.service.complete(
            self, action, variables=variables, comment=comment)
        self.load(updated_task)

    def delegate(self, actors, comment=None):
        # type: (Text, Optional[Text]) -> None
        """ Delegate the task to someone else. """
        self.service.transfer(self, 'delegate', actors, comment=comment)
        self.load()

    def reassign(self, actors, comment=None):
        # type: (Text, Optional[Text]) -> None
        """ Reassign the task to someone else. """
        self.service.transfer(self, 'reassign', actors, comment=comment)
        self.load()


class User(RefreshableModel):
    """ User. """
    _valid_properties = {
        'entity-type': 'user',
        'id': None,
        'properties': {},
        'extendedGroups': [],
        'isAdministrator': False,
        'isAnonymous': False,
    }
    service = None  # type: UsersAPI

    def __init__(self, **kwargs):
        # type: (Any) -> None
        super(User, self).__init__(**kwargs)
        for key, default in User._valid_properties.items():
            key = key.replace('-', '_')
            setattr(self, key, kwargs.get(key, default))

    @property
    def uid(self):
        # type: () -> Text
        return self.id

    def change_password(self, password):
        # type: (Text) -> None
        """
        Change user password.

        :param password: New password to set
        """
        self.properties['password'] = password
        self.save()

    def delete(self):
        # type: () -> None
        """ Delete the user. """
        self.service.delete(self.uid)


class Workflow(Model):
    """ Workflow. """
    _valid_properties = {
        'entity-type': 'workflow',
        'id': None,
        'name': None,
        'title': None,
        'state': None,
        'workflowModelName': None,
        'initiator': None,
        'attachedDocumentIds': [],
        'variables': {},
        'graphResource': None,
    }
    service = None  # type: WorkflowsAPI

    def __init__(self, **kwargs):
        # type: (Any) -> None
        super(Workflow, self).__init__(**kwargs)
        for key, default in Workflow._valid_properties.items():
            key = key.replace('-', '_')
            setattr(self, key, kwargs.get(key, default))

    @property
    def tasks(self):
        # type: () -> List[Task]
        """ Return the tasks associated with the workflow. """
        return self.service.tasks(self)

    @property
    def uid(self):
        # type: () -> Text
        return self.id

    def delete(self):
        # type: () -> None
        """ Delete the workflow. """
        self.service.delete(self.uid)

    def graph(self):
        # type: () -> Dict[Text, Any]
        """ Return a JSON representation of the workflow graph. """
        return self.service.graph(self)



# coding: utf-8
from __future__ import unicode_literals

from .compat import get_bytes, quote, text
from .constants import CHUNK_LIMIT, MAX_RETRY, UPLOAD_CHUNK_SIZE
from .endpoint import APIEndpoint
from .exceptions import UploadError
from .models import Batch, Blob
from .utils import SwapAttr

try:
    from typing import TYPE_CHECKING
    if TYPE_CHECKING:
        from typing import Any, Dict, List, Optional, Text, Tuple, Union
        from .client import NuxeoClient
        OptInt = Optional[int]
except ImportError:
    pass


class API(APIEndpoint):
    """ Endpoint for uploads. """
    def __init__(self, client, endpoint='upload', headers=None):
        # type: (NuxeoClient, Text, Optional[Dict[Text, Text]]) -> None
        super(API, self).__init__(
            client, endpoint=endpoint, cls=Blob, headers=headers)

    def get(self, batch_id, file_idx=None):
        # type: (Text, Optional[int]) -> Union[List[Blob], Blob]
        """
        Get the detail of a batch.

        If file_idx is None, returns the details of all its blobs,
        otherwise returns the details of the corresponding blob.

        :param batch_id: the id of the batch
        :param file_idx: the index of the blob
        :return: the batch details
        """
        path = batch_id
        if file_idx is not None:
            path = '{}/{}'.format(path, file_idx)

        resource = super(API, self).get(path=path)

        if file_idx is not None:
            resource.batch_id = batch_id
            resource.fileIdx = file_idx
        elif not resource:
            return []
        return resource

    def post(self):
        # type: () -> Batch
        """
        Create a batch.

        :return: the created batch
        """
        response = self.client.request('POST', self.endpoint)
        return Batch.parse(response.json(), service=self)

    batch = post  # Alias for clarity

    def put(self, **kwargs):
        # type: (Any) -> None
        raise NotImplementedError()

    def delete(self, batch_id, file_idx=None):
        # type: (Text, Optional[int]) -> None
        """
        Delete a batch or a blob.

        If the file_idx is None, deletes the batch,
        otherwise deletes the corresponding blob.

        :param batch_id: the id of the batch
        :param file_idx: the index of the blob
        """
        if file_idx is not None:
            target = '{}/{}'.format(batch_id, file_idx)
            super(API, self).delete(target)
        else:
            target = batch_id
            with SwapAttr(self, '_cls', Batch):
                super(API, self).delete(target)

    def send_data(
        self,
        name,  # type: Text
        data,  # type: Union[Text, bytes]
        path,  # type: Text
        chunked,  # type: bool
        index,  # type: int
        headers,  # type: Dict[Text, Text]
    ):
        # type: (...) -> Blob
        """
        Send data/chunks to the server.


        :param name: name of the file being uploaded
        :param data: data being sent
        :param path: url for the upload
        :param chunked: True if the upload is in chunks
        :param index: which chunk is being sent (0 if not chunked)
        :param headers: HTTP request headers
        :return: the blob info
        """
        if chunked:
            headers['X-Upload-Chunk-Index'] = text(index)

        for i in range(MAX_RETRY):
            response = super(API, self).post(
                resource=data,
                path=path,
                raw=True,
                headers=headers,
                default={}
            )
            if response:
                break
        else:
            chunk = index if chunked else None
            raise UploadError(name, chunk=chunk)
        return response

    def state(self, path, blob):
        # type: (Text, Blob) -> Tuple[OptInt, OptInt, OptInt, Blob]
        """
        Get the state of a blob.

        If the blob upload has not begun yet, the server
        will return a 404 error, so we initialize the
        different values.
        If the blob upload is incomplete, we return the
        values the server sent us.
        If the blob upload is complete, we return None
        for these values.

        :param path: path for the request
        :param blob: the target blob
        :return: the chunk size, chunk count, the index
                 of the next blob to upload, and the
                 response from the server
        """
        info = super(API, self).get(path, default=None)

        if info:
            chunk_count = int(info.chunkCount)
            chunk_size = int(info.uploadedSize)
            index = int(info.uploadedChunkIds[-1]) + 1
        else:  # It's a new upload
            chunk_size = UPLOAD_CHUNK_SIZE
            chunk_count = (blob.size // chunk_size +
                           (blob.size % chunk_size > 0))
            index = 0

        return chunk_size, chunk_count, index, info

    def upload(self, batch, blob, chunked=False, limit=CHUNK_LIMIT):
        # type: (Batch, Blob, bool, int) -> Blob
        """
        Upload a blob.

        Can be used to upload a new blob or resume
        the upload of a chunked blob.

        :param batch: batch of the upload
        :param blob: blob to upload
        :param chunked: if True, send in chunks
        :param limit: if blob is bigger, send in chunks
        :return: uploaded blob details
        """
        chunked = (chunked or blob.size > limit) and blob.size > 0
        response = None

        headers = self.headers
        headers.update({
            'Cache-Control': 'no-cache',
            'X-File-Name': quote(get_bytes(blob.name)),
            'X-File-Size': text(blob.size),
            'X-File-Type': blob.mimetype,
            'Content-Length': text(blob.size),
        })

        path = '{}/{}'.format(batch.batchId, batch._upload_idx)

        if chunked:
            chunk_size, chunk_count, index, info = self.state(path, blob)

            headers.update({
                'X-Upload-Type': 'chunked',
                'X-Upload-Chunk-Count': text(chunk_count),
                'Content-Length': text(chunk_size)
            })
        else:
            chunk_size, chunk_count, index = blob.size or None, 1, 0

        with blob as source:
            if index:
                source.seek(index * chunk_size)
            while index < chunk_count:
                data = source.read(chunk_size) if chunk_size else source.read()
                response = self.send_data(
                    blob.name, data, path, chunked, index, headers)
                index += 1

        response.batch_id = batch.uid
        return response

    def execute(self, batch, operation, file_idx=None, params=None):
        # type: (Batch, Text, Optional[int], Optional[Dict[Text,Any]]) -> Any
        """
        Execute an operation with the batch or one of its files as an input.

        :param batch: input for the operation
        :param operation: operation to execute
        :param file_idx: if not None, sole input of the operation
        :param params: parameters for the operation
        :return: the output of the operation
        """
        path = '{}/{}'.format(self.endpoint, batch.uid)
        if file_idx is not None:
            path = '{}/{}'.format(path, file_idx)

        path = '{}/execute/{}'.format(path, operation)

        return self.client.request('POST', path, data={'params': params})

    def attach(self, batch, doc, file_idx=None):
        # type: (Batch, Text, Optional[int]) -> Any
        """
        Attach one or all files of a batch to a document.

        :param batch: batch to attach
        :param doc: document to attach
        :param file_idx: if not None, only this file will be attached
        :return: the output of the attach operation
        """
        params = {'document': doc}
        if file_idx is None and batch._upload_idx > 1:
            params['xpath'] = 'files:files'
        return self.execute(batch, 'Blob.Attach', file_idx, params)



# coding: utf-8
from __future__ import unicode_literals

from .endpoint import APIEndpoint
from .exceptions import BadQuery, HTTPError, UnavailableConvertor
from .models import Document
from .utils import SwapAttr
from .workflows import API as WorkflowsAPI

try:
    from typing import TYPE_CHECKING
    if TYPE_CHECKING:
        from typing import Any, Dict, List, Optional, Text, Union
        from .client import NuxeoClient
        from .models import Blob, Workflow
        from .operations import API as OperationsAPI
except ImportError:
    pass


class API(APIEndpoint):
    """ Endpoint for documents. """
    def __init__(
        self,
        client,  # type: NuxeoClient
        operations,  # type: OperationsAPI
        workflows,  # type: WorkflowsAPI
        endpoint=None,  # type: Text
        headers=None,  # type: Optional[Dict[Text, Text]]
    ):
        # type: (...) -> None
        self.operations = operations
        self.workflows_api = workflows
        super(API, self).__init__(
            client, endpoint=endpoint, cls=Document, headers=headers)

    def get(self, uid=None, path=None):
        # type: (Optional[Text], Optional[Text]) -> Document
        """
        Get the detail of a document.

        :param uid: the uid of the document
        :param path: the path of the document
        :return: the document
        """
        return super(API, self).get(path=self._path(uid=uid, path=path))

    def post(self, document, parent_id=None, parent_path=None):
        # type: (Document, Optional[Text], Optional[Text]) -> Document
        """
        Create a document.

        :param document: the document to create
        :param parent_id: the id of the parent document
        :param parent_path: the path of the parent document
        :return: the created document
        """
        return super(API, self).post(
            document, path=self._path(uid=parent_id, path=parent_path))

    create = post  # Alias for clarity

    def put(self, document):
        # type: (Document) -> Document
        """
        Update a document.

        :param document: the document to update
        :return: the updated document
        """
        return super(API, self).put(
            document, path=self._path(uid=document.uid))

    def delete(self, document_id):
        # type: (Text) -> None
        """
        Delete a document.

        :param document_id: the id of the document to delete
        """
        super(API, self).delete(self._path(uid=document_id))

    def exists(self, uid=None, path=None):
        # type: (Optional[Text], Optional[Text]) -> bool
        """
        Check if a document exists.

        :param uid: the id of the document to check
        :param path: the path of the document to check
        :return: True if it exists, else False
        """
        try:
            self.get(uid=uid, path=path)
            return True
        except HTTPError as e:
            if e.status != 404:
                raise e
        return False

    def add_permission(self, uid, params):
        # type: (Text, Dict[Text, Any]) -> None
        """
        Add a permission to a document.

        :param uid: the uid of the document
        :param params: the permissions to add
        """
        self.operations.execute(
            command='Document.AddPermission', input_obj=uid, params=params)

    def convert(self, uid, options):
        # type: (Text, Dict[Text, Text]) -> Union[Text, Dict[Text, Any]]
        """
        Convert a blob into another format.

        :param uid: the uid of the blob to be converted
        :param options: the target type, target format,
                        or converter for the blob
        :return: the response from the server
        """
        xpath = options.pop('xpath', 'blobholder:0')
        adapter = 'blob/{}/@convert'.format(xpath)
        if ('converter' not in options
                and 'type' not in options
                and'format' not in options):
            raise BadQuery(
                'One of (converter, type, format) is mandatory in options')

        try:
            return super(API, self).get(
                path=self._path(uid=uid), params=options,
                adapter=adapter, raw=True)
        except HTTPError as e:
            if 'is not registered' in e.message:
                raise BadQuery(e.message)
            if 'is not available' in e.message:
                raise UnavailableConvertor(options)
            raise e

    def fetch_acls(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """
        Fetch the ACLs of a document.

        :param uid: the uid of the document
        :return: the ACLs
        """
        req = super(API, self).get(
            path=self._path(uid=uid), cls=dict, headers=self.headers,
            enrichers=['acls'])
        return req['contextParameters']['acls']

    def fetch_audit(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """
        Fetch the audit of a document.

        :param uid: the uid of the document
        :return: the audit
        """
        return super(API, self).get(
            self._path(uid=uid), adapter='audit', cls=dict)

    def fetch_lock_status(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """
        Fetch the lock status of a document.

        :param uid: the uid of the document
        :return: the lock status
        """
        headers = self.headers or {}
        headers.update({'fetch-document': 'lock'})
        req = super(API, self).get(
            path=self._path(uid=uid), cls=dict, headers=headers)
        if 'lockOwner' in req:
            return {
                'lockCreated': req['lockOwner'],
                'lockOwner': req['lockOwner']
            }
        else:
            return {}

    def fetch_rendition(self, uid, name):
        # type: (Text, Text) -> Union[Text, bytes]
        """
        Fetch a rendition of a document.

        :param uid: the uid of the document
        :param name: the name of the rendition
        :return: the corresponding rendition
        """
        adapter = 'rendition/{}'.format(name)
        return super(API, self).get(
            path=self._path(uid=uid), raw=True, adapter=adapter)

    def fetch_renditions(self, uid):
        # type: (Text) -> List[Union[Text, bytes]]
        """
        Fetch all renditions of a document.

        :param uid: the uid of a document
        :return: the renditions
        """
        headers = self.headers or {}
        headers.update({'enrichers-document': 'renditions'})

        req = super(API, self).get(
            path=self._path(uid=uid), cls=dict, headers=headers)
        return [rend['name']
                for rend in req['contextParameters']['renditions']]

    def follow_transition(self, uid, name):
        # type: (Text, Text) -> Dict[Text, Any]
        """
        Follow a lifecycle transition.

        :param uid: the uid of the target document
        :param name: the name of the transition
        """
        params = {'value': name}
        return self.operations.execute(
            command='Document.FollowLifecycleTransition',
            input_obj=uid, params=params)

    def fetch_blob(self, uid=None, path=None, xpath='blobholder:0'):
        # type: (Optional[Text], Optional[Text], Text) -> Blob
        """
        Get the blob of a document.

        :param uid: the uid of the document
        :param path: the path of the document
        :param xpath: the xpath of the blob
        :return: the blob
        """
        adapter = 'blob/{}'.format(xpath)
        return super(API, self).get(
            path=self._path(uid=uid, path=path), raw=True, adapter=adapter)

    def get_children(self, uid=None, path=None):
        # type: (Optional[Text], Optional[Text]) -> List[Document]
        """
        Get the children of a document.

        :param uid: the uid of the document
        :param path: the path of the document
        :return: the document children
        """
        return super(API, self).get(
            path=self._path(uid=uid, path=path), adapter='children')

    def has_permission(self, uid, permission):
        # type: (Text, Text) -> bool
        """
        Check if a document has a permission.

        :param uid: the uid of the document
        :param permission: the permission to check
        :return: True if the document has it, False otherwise
        """
        req = super(API, self).get(
            path=self._path(uid=uid), cls=dict, headers=self.headers,
            enrichers=['permissions'])
        return permission in req['contextParameters']['permissions']

    def lock(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """ Lock a document. """
        return self.operations.execute(
            command='Document.Lock', input_obj=uid)

    def move(self, uid, dst, name=None):
        # type: (Text, Text, Optional[Text]) -> Dict[Text, Any]
        """
        Move a document and eventually rename it.

        :param uid: the uid of the target document
        :param dst: the destination
        :param name: the new name
        """
        params = {'target': dst}
        if name:
            params['name'] = name
        return self.operations.execute(
            command='Document.Move', input_obj=uid, params=params)

    def query(self, opts=None):
        # type: (Optional[Dict[Text, Text]]) -> Dict[Text, Any]
        """
        Run a query on the documents.

        :param opts: a query or a pageProvider
        :return: the corresponding documents
        """
        opts = opts or {}
        if 'query' in opts:
            query = 'NXQL'
        elif 'pageProvider' in opts:
            query = opts['pageProvider']
        else:
            raise BadQuery('Need either a pageProvider or a query')

        path = 'query/{}'.format(query)
        res = super(API, self).get(path=path, params=opts, cls=dict)
        res['entries'] = [Document.parse(entry, service=self)
                          for entry in res['entries']]
        return res

    def remove_permission(self, uid, params):
        # type: (Text, Dict[Text, Text]) -> None
        """
        Remove a permission on a document.

        :param uid: the uid of the document
        :param params: the permission to remove
        """
        self.operations.execute(
            command='Document.RemovePermission', input_obj=uid, params=params)

    def trash(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """
        Trash the document.

        :param uid: the uid of the document
        """
        return self.operations.execute(
            command='Document.Trash', input_obj=uid)

    def unlock(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """ Unlock a document. """
        return self.operations.execute(
            command='Document.Unlock', input_obj=uid)

    def untrash(self, uid):
        # type: (Text) -> Dict[Text, Any]
        """
        Untrash the document.

        :param uid: the uid of the document
        """
        return self.operations.execute(
            command='Document.Untrash', input_obj=uid)

    def workflows(self, document):
        # type: (Document) -> Union[Workflow, List[Workflow]]
        """ Get the workflows of a document. """
        path = 'id/{}/@workflow'.format(document.uid)

        with SwapAttr(self.workflows_api, 'endpoint', self.endpoint):
            return super(WorkflowsAPI, self.workflows_api).get(path=path)

    def _path(self, uid=None, path=None):
        # type: (Optional[Text], Optional[Text]) -> Text
        if uid:
            path = 'repo/{}/id/{}'.format(self.client.repository, uid)
        elif path:
            path = 'repo/{}/path{}'.format(self.client.repository, path)
        return path



# coding: utf-8
from __future__ import unicode_literals

from .compat import text

try:
    from typing import TYPE_CHECKING
    if TYPE_CHECKING:
        from typing import Any, Dict, Optional, Text
except ImportError:
    pass


class NuxeoError(Exception):
    """ Mother class for all exceptions. """


class BadQuery(NuxeoError):
    """
    Exception thrown either by an operation failure:
        - when the command is not valid
        - on unexpected parameter
        - on missing required parameter
        - when a parameter has not the required type
    Or by any requests to the server with invalid data.
    """


class CorruptedFile(NuxeoError):
    """ Exception thrown when digests of a downloaded blob are different. """

    def __init__(self, filename, server_digest, local_digest):
        # type: (Text, Text, Text) -> None
        self.filename = filename
        self.server_digest = server_digest
        self.local_digest = local_digest

    def __repr__(self):
        # type: () -> Text
        err = ('CorruptedFile {!r}: server digest '
               'is {!r}, local digest is {!r}')
        return err.format(self.filename, self.server_digest, self.local_digest)

    def __str__(self):
        # type: () -> Text
        return repr(self)


class HTTPError(NuxeoError):
    """ Exception thrown when the server returns an error. """
    _valid_properties = {
        'status': None,
        'message': None,
        'stacktrace': None
    }

    def __init__(self, **kwargs):
        # type: (Any) -> None
        for key, default in HTTPError._valid_properties.items():
            setattr(self, key, kwargs.get(key, default))

    def __repr__(self):
        # type: () -> Text
        return '%s(%d), error: %s, server trace: %s' % (
            type(self).__name__, self.status, self.message, self.stacktrace)

    def __str__(self):
        # type: () -> Text
        return repr(self)

    @classmethod
    def parse(cls, json):
        # type: (Dict[Text, Any]) -> HTTPError
        """ Parse a JSON object into a model instance. """
        model = cls()

        for key, val in json.items():
            if key in cls._valid_properties:
                setattr(model, key, val)
        return model


class InvalidBatch(NuxeoError):
    """ Exception thrown when accessing inexistant or deleted batches. """


class Unauthorized(HTTPError):
    """ Exception thrown when the HTTPError code is 401 or 403. """


class UnavailableConvertor(NuxeoError):
    """
    Exception thrown when a converter is registered but not
    available right now (e.g. not installed on the server).

    :param options: options passed to the conversion request
    """
    def __init__(self, options):
        # type: (Dict[Text, Any]) -> None
        self.options = options
        self.message = text(self)

    def __repr__(self):
        # type: () -> Text
        return ('UnavailableConvertor: conversion with options {!r}'
                ' is not available').format(self.options)

    def __str__(self):
        # type: () -> Text
        return repr(self)


class UploadError(NuxeoError):
    """
    Exception thrown when an upload fails even after retries.
    """
    def __init__(self, name, chunk=None):
        # type: (Text, Optional[int]) -> None
        self.name = name
        self.chunk = chunk

    def __repr__(self):
        # type: () -> Text
        err = 'UploadError: unable to upload file {!r}'.format(self.name)
        if self.chunk:
            err = '{} (failed at chunk {})'.format(err, self.chunk)
        return err

    def __str__(self):
        # type: () -> Text
        return repr(self)



# coding: utf-8
from __future__ import unicode_literals

import atexit
import json
import logging

import requests

from . import (__version__, directories, documents, groups,
               operations, tasks, uploads, users, workflows)
from .auth import TokenAuth
from .compat import text
from .constants import (CHUNK_SIZE, DEFAULT_API_PATH, DEFAULT_APP_NAME,
                        DEFAULT_URL)
from .exceptions import BadQuery, HTTPError, Unauthorized
from .utils import json_helper

try:
    from typing import TYPE_CHECKING
    if TYPE_CHECKING:
        from typing import Any, Dict, Optional, Text, Tuple, Type, Union
        from requests.auth import AuthBase
        AuthType = Optional[Union[Tuple[Text, Text], AuthBase]]
except ImportError:
    pass

logger = logging.getLogger(__name__)


class NuxeoClient(object):
    """
    The HTTP client used by Nuxeo.

    :param auth: An authentication object passed to Requests
    :param host: The url of the Nuxeo Platform
    :param api_path: The API path appended to the host url
    :param chunk_size: The size of the chunks for blob download
    :param kwargs: kwargs passed to :func:`NuxeoClient.request`
    """

    def __init__(
        self,
        auth=None,  # type: AuthType
        host=DEFAULT_URL,  # type: Text
        api_path=DEFAULT_API_PATH,  # type: Text
        chunk_size=CHUNK_SIZE,  # type: int
        **kwargs  # type: Any
    ):
        # type: (...) -> None
        self.auth = auth
        self.host = host
        self.api_path = api_path
        self.chunk_size = chunk_size

        version = kwargs.pop('version', '')
        app_name = kwargs.pop('app_name', DEFAULT_APP_NAME)
        self.headers = {
            'X-Application-Name': app_name,
            'X-Client-Version': version,
            'User-Agent': app_name + '/' + version,
            'Accept': 'application/json, */*'
        }
        self.schemas = kwargs.get('schemas', '*')
        self.repository = kwargs.pop('repository', 'default')
        self._session = requests.session()
        cookies = kwargs.pop('cookies', None)
        if cookies:
            self._session.cookies = cookies
        self._session.stream = True
        self.client_kwargs = kwargs
        atexit.register(self.on_exit)

        # Ensure the host is well formatted
        if not self.host.endswith('/'):
            self.host += '/'

    def __repr__(self):
        # type: () -> Text
        fmt = '{name}<host={cls.host!r}, version={cls.server_version!r}>'
        return fmt.format(name=type(self).__name__, cls=self)

    def __str__(self):
        # type: () -> Text
        return repr(self)

    def on_exit(self):
        # type: () -> None
        self._session.close()

    def query(
        self,
        query,  # type: Text
        params=None,  # type: Dict[Text, Any]
    ):
        """
        Query the server with the specified NXQL query.
        Additional qery parameters can be set via the `params` argument:

            >>> nuxeo.client.query('SQL query', params={'properties': '*'})

        You can find what parameters to tweak under the `Repository.Query`
        operation details.
        """

        data = {'query': query}
        if params:
            data.update(params)

        url = self.api_path + '/search/lang/NXQL/execute'
        return self.request('GET', url, params=data).json()

    def set(self, repository=None, schemas=None):
        # type: (Optional[Text], Optional[Text]) -> NuxeoClient
        """
        Set the repository and/or the schemas for the requests.

        :return: The client instance after adding the settings
        """
        if repository:
            self.repository = repository

        if schemas:
            if isinstance(schemas, list):
                schemas = ','.join(schemas)
            self.schemas = schemas

        return self

    def request(
        self,
        method,  # type: Text
        path,  # type: Text
        headers=None,  # type: Optional[Dict[Text, Text]]
        data=None,  # type: Optional[Any]
        raw=False,  # type: bool
        **kwargs  # type: Any
    ):
        # type: (...) -> Union[requests.Response, Any]
        """
        Send a request to the Nuxeo server.

        :param method: the HTTP method
        :param path: the path to append to the host
        :param headers: the headers for the HTTP request
        :param data: data to put in the body
        :param raw: if True, don't parse the data to JSON
        :param kwargs: other parameters accepted by
               :func:`requests.request`
        :return: the HTTP response
        """
        if method not in ('GET', 'HEAD', 'POST', 'PUT',
                          'DELETE', 'CONNECT', 'OPTIONS', 'TRACE'):
            raise BadQuery('method parameter is not a valid HTTP method.')

        # Construct the full URL without double slashes
        url = self.host + path.lstrip('/')
        if 'adapter' in kwargs:
            url = '{}/@{}'.format(url, kwargs.pop('adapter'))

        kwargs.update(self.client_kwargs)

        headers = headers or {}
        if 'Content-Type' not in headers:
            headers['Content-Type'] = kwargs.pop(
                'content_type', 'application/json')
        headers.update({
            'X-NXDocumentProperties': self.schemas,
            'X-NXRepository': self.repository
        })
        enrichers = kwargs.pop('enrichers', None)
        if enrichers:
            headers['X-NXenrichers.document'] = ', '.join(enrichers)

        headers.update(self.headers)

        if data and not isinstance(data, bytes) and not raw:
            data = json.dumps(data, default=json_helper)

        # Set the default value to `object` to allow someone
        # to set `default` to `None`.
        default = kwargs.pop('default', object)

        logger.debug(
            ('Calling {!r} with headers={!r}, '
             'params={!r} and cookies={!r}').format(
                url, headers, kwargs.get('params', data if not raw else {}),
                self._session.cookies))

        try:
            resp = self._session.request(
                method, url, headers=headers,
                auth=self.auth, data=data, **kwargs)
            resp.raise_for_status()
        except Exception as exc:
            if default is object:
                raise self._handle_error(exc)
            resp = default
        else:
            content_size = resp.headers.get('content-length', self.chunk_size)
            if int(content_size) <= self.chunk_size:
                content = resp.text
            else:
                content = '<Too much data to display>'
            logger.debug('Response from {!r}: {!r} with cookies {!r}'.format(
                url, content, self._session.cookies))

        return resp

    def request_auth_token(
        self,
        device_id,  # type: Text
        permission,  # type: Text
        app_name=DEFAULT_APP_NAME,  # type: Text
        device=None,  # type: Optional[Text]
        revoke=False,  # type: bool
    ):
        # type: (...) -> Text
        """
        Request a token for the user.

        :param device_id: device identifier
        :param permission: read/write permissions
        :param app_name: application name
        :param device: optional device description
        :param revoke: revoke the token
        """

        parameters = {
            'deviceId': device_id,
            'applicationName': app_name,
            'permission': permission,
            'revoke': text(revoke).lower(),
        }
        if device:
            parameters['deviceDescription'] = device

        path = 'authentication/token'
        token = self.request('GET', path, params=parameters).text

        # Use the (potentially re-newed) token from now on
        if not revoke:
            self.auth = TokenAuth(token)
        return token

    def is_reachable(self):
        # type: () -> bool
        """ Check if the Nuxeo Platform is reachable. """
        response = self.request('GET', 'runningstatus', default=False)
        if isinstance(response, requests.Response):
            return response.ok
        else:
            return bool(response)

    def server_info(self, force=False):
        # type: (bool) -> Dict[Text, Text]
        """
        Retreive server information.

        :param bool force: Force information renewal.
        """

        if force or not getattr(self, '_server_info', None):
            response = self.request('GET', 'json/cmis', default={})
            if isinstance(response, requests.Response):
                info = response.json()['default']
            else:
                info = response
            setattr(self, '_server_info', info)
        return self._server_info

    @property
    def server_version(self):
        # type: () -> Text
        """ Return the server version. """
        return self.server_info().get('productVersion', '')

    @staticmethod
    def _handle_error(error):
        # type: (Exception) -> Exception
        """
        Log error and handle raise.

        :param error: The error to handle
        """
        if isinstance(error, requests.HTTPError):
            try:
                error_data = error.response.json()
            except ValueError:
                error_data = {'status': error.response.status_code,
                              'message': error.response.content}

            error_class = (Unauthorized
                           if error_data.get('status') in (401, 403)
                           else HTTPError)

            error = error_class.parse(error_data)
        return error


class Nuxeo(object):
    """
    Instantiate the client and all the API Endpoints.

    :param auth: the authenticator
    :param host: the host URL
    :param app_name: the name of the application using the client
    :param client: the client class
    :param kwargs: any other argument to forward to every requests calls
    """
    def __init__(
        self,
        auth=None,  # type: Optional[Tuple[Text, Text]]
        host=DEFAULT_URL,  # type: Text
        app_name=DEFAULT_APP_NAME,  # type: Text
        version=__version__,  # type: Text
        client=NuxeoClient,  # type: Type[NuxeoClient]
        **kwargs  # type: Any
    ):
        # type: (...) -> None
        if requests.__version__ < '2.12.2':
            from warnings import warn
            warn('Requests >= 2.12.2 required for auth unicode support.')

        self.client = client(auth, host=host, app_name=app_name,
                             version=version, **kwargs)
        self.operations = operations.API(self.client)
        self.directories = directories.API(self.client)
        self.groups = groups.API(self.client)
        self.tasks = tasks.API(self.client)
        self.uploads = uploads.API(self.client)
        self.users = users.API(self.client)
        self.workflows = workflows.API(self.client, self.tasks)
        self.documents = documents.API(
            self.client, self.operations, self.workflows)



# coding: utf-8
from __future__ import unicode_literals

import logging
import mimetypes
import sys

import hashlib

try:
    from typing import TYPE_CHECKING
    if TYPE_CHECKING:
        from _hashlib import HASH
        from typing import Any, Dict, Optional, Text, Type, Union
except ImportError:
    pass


logger = logging.getLogger(__name__)
WIN32_PATCHED_MIME_TYPES = {
    'image/pjpeg': 'image/jpeg',
    'image/x-png': 'image/png',
    'image/bmp': 'image/x-ms-bmp',
    'audio/x-mpg': 'audio/mpeg',
    'video/x-mpeg2a': 'video/mpeg',
    'application/x-javascript': 'application/javascript',
    'application/x-msexcel': 'application/vnd.ms-excel',
    'application/x-mspowerpoint': 'application/vnd.ms-powerpoint',
    'application/x-mspowerpoint.12':
    'application/vnd.openxmlformats-officedocument.presentationml.presentation'
}


def get_digest_algorithm(digest):
    # type: (Text) -> Optional[Text]

    # Available algorithms
    digesters = {
        32: 'md5',
        40: 'sha1',
        56: 'sha224',
        64: 'sha256',
        96: 'sha384',
        128: 'sha512',
    }

    # Ensure the digest is hexadecimal
    try:
        int(digest, 16) >= 0
    except (TypeError, ValueError):
        return None

    return digesters.get(len(digest), None)


def get_digest_hash(algorithm):
    # type: (Text) -> Optional[HASH]

    # Retrieve the hashlib function for the given digest, None if not found
    func = getattr(hashlib, algorithm, None)

    return func() if func else None


def get_digester(digest):
    # type: (Text) -> Optional[HASH]
    """
    Get the digester corresponding to the given hash.

    To choose the digester used by the server, see
    https://doc.nuxeo.com/nxdoc/file-storage-configuration/#configuring-the-default-blobprovider

    :param digest: the hash
    :return: the digester function
    """

    algo = get_digest_algorithm(digest)
    func = get_digest_hash(algo) if algo else None

    if not func:
        logger.error('No valid hash algorithm found for digest %r', digest)

    return func


def guess_mimetype(filename):
    # type: (Text) -> Text
    """ Guess the mimetype of a given file. """
    mime_type, _ = mimetypes.guess_type(filename)
    if mime_type:
        if sys.platform == 'win32':
            # Patch bad Windows MIME types
            # See http://bugs.python.org/issue15207 and NXP-11660
            mime_type = WIN32_PATCHED_MIME_TYPES.get(mime_type, mime_type)
        return mime_type

    return 'application/octet-stream'


def json_helper(obj):
    # type: (Any) -> Dict[Text, Any]
    return obj.to_json()


class SwapAttr(object):
    """
    Context manager to swap an attribute's value:

        >>> # self.person equals 'Alice'
        >>> with SwapAttr(self, 'person', 'Bob'):
        ...     # ...

    """

    def __init__(self, obj, attr, value):
        # type: (Any, Text, Any) -> None
        self.obj = obj
        self.attr = attr
        self.value = value
        self.old_value = getattr(obj, attr)

    def __enter__(self):
        # type: () -> None
        setattr(self.obj, self.attr, self.value)

    def __exit__(self, *args):
        # type: (Any) -> None
        setattr(self.obj, self.attr, self.old_value)



# coding: utf-8
"""
Nuxeo REST API Python client.
https://doc.nuxeo.com/nxdoc/rest-api/

You can always get the latest version of this module at:
    https://github.com/nuxeo/nuxeo-python-client
If that URL should fail, try contacting the author.

Contributors:
    Antoine Taillefer <ataillefer@nuxeo.com>
    Rémi Cattiau
    Mickaël Schoentgen <mschoentgen@nuxeo.com>
    Léa Klein <lklein@nuxeo.com>
    and https://github.com/nuxeo/nuxeo-python-client/graphs/contributors
"""
from __future__ import unicode_literals


__author__ = 'Nuxeo'
__version__ = '2.0.4'
__copyright__ = """
    Copyright Nuxeo (https://www.nuxeo.com) and others.

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""
